#!/usr/local/bin/ruby -w
#
# vim:sw=2 ts=8:et sta:fdm=marker
#
#
# Copyright (c) 2006 Ariff Abdullah <ariff@FreeBSD.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#        $MyBSD$
#
# Date: Tue Aug 22 23:09:20 MYT 2006
#   OS: FreeBSD yuki.MyBSD.org.my 6.1-STABLE i386
#

PIN_OUT = 1 << 0
PIN_IN = 1 << 1

SOUND_MIXER_NRDEVICES = 25
SOUND_MIXER_VOLUME    = 0
SOUND_MIXER_BASS      = 1
SOUND_MIXER_TREBLE    = 2
SOUND_MIXER_SYNTH     = 3
SOUND_MIXER_PCM       = 4
SOUND_MIXER_SPEAKER   = 5
SOUND_MIXER_LINE      = 6
SOUND_MIXER_MIC       = 7
SOUND_MIXER_CD        = 8
SOUND_MIXER_IMIX      = 9
SOUND_MIXER_ALTPCM    = 10
SOUND_MIXER_RECLEV    = 11
SOUND_MIXER_IGAIN     = 12
SOUND_MIXER_OGAIN     = 13
SOUND_MIXER_LINE1     = 14
SOUND_MIXER_LINE2     = 15
SOUND_MIXER_LINE3     = 16
SOUND_MIXER_DIGITAL1  = 17
SOUND_MIXER_DIGITAL2  = 18
SOUND_MIXER_DIGITAL3  = 19
SOUND_MIXER_PHONEIN   = 20
SOUND_MIXER_PHONEOUT  = 21
SOUND_MIXER_VIDEO     = 22
SOUND_MIXER_RADIO     = 23
SOUND_MIXER_MONITOR   = 24

HDA_CODECS = {
  0x10ec0260 => 'Realtek ALC260',
  0x10ec0861 => 'Realtek ALC861',
  0x10ec0880 => 'Realtek ALC880',
  0x10ec0882 => 'Realtek ALC882',
  0x11d41981 => 'Analog Device AD1981HD',
  0x11d41983 => 'Analog Device AD1983',
  0x11d41986 => 'Analog Device AD1986A',
  0x434d4980 => 'CMedia CMI9880',
  0x83847680 => 'Sigmatel STAC9221',
  0x83847683 => 'Sigmatel STAC9221D',
  0x83847690 => 'Sigmatel STAC9220',
  0x14f15045 => 'Conexant Venice',
  0x14f15047 => 'Conexant Waikiki',
}
HDA_CODECS.default = 'Unknown'

OUT_PATH     = 1 << 0
IN_PATH      = 1 << 1
REC_SELECTOR = 1 << 2

OSS_LABELS = [
	"vol", "bass", "treble", "synth", "pcm", "speaker", "line",
	"mic", "cd", "mix", "pcm2", "rec", "igain", "ogain",
	"line1", "line2", "line3", "dig1", "dig2", "dig3",
	"phin", "phout", "video", "radio", "monitor"
]

MAXDEPTH = 25

CTL_OUT = 0x1
CTL_IN  = 0x2

class Amp
  def initialize(nid = -1, index = -1, dir = 0, mute = 0, step = 0, size = 0, off = 0)
    @mute = mute
    @step = step
    @size = size
    @off = off
    @dir = dir
    @nid = nid
    @index = index
    @flags = 0
  end
  attr_accessor :dir, :mute, :step, :size, :off, :nid, :index, :flags
end

class Widget
  def initialize(str = '(unknown)', nid = 0, mode = :ANALOG, conns = [])
    @nid = nid
    @conns = conns
    @mode = mode
    @str = str.dup()
    @valid = false
    @selconn = -1
    @parsed = 0
    @ctlflags = 0
    @pindir = 0
  end
  def nconns
    @conns ||= []
    @conns.size()
  end
  attr_accessor :str, :nid, :conns, :mode, :valid, :selconn, :parsed, :pindir, :ctlflags
end

class Pin < Widget
end

class Mixer < Widget
end

class Input < Widget
end

class Output < Widget
end

class Beep < Widget
end

class Vendor < Widget
end

class Selector < Widget
end

class Volume < Widget
end

class Power < Widget
end

class HDAC
  attr_accessor :vendor, :device, :revision, :stepping, :pci_subvendor, :codec_name
end

def hdac_widget_get(nodes, nid)
  nodes.each do |n| return n if n.nid == nid end
  nil
end

nodes = []
amps = []

nid = nil
name = nil
conns = nil
mode = nil
wclass = nil
ampdir = nil
outamp = nil
inamp = nil
pindir = nil
hdac = nil

dumpfile = ARGV[0] || '/HDA_NODES2.TXT'
File.foreach(dumpfile) do |l|
  next unless /^pcm\d+:\s+/ =~ l
  l.chomp!
  l.sub!(/^pcm\d+:\s*/, '')
  l.sub!(/\s*$/, '')
  if /^nid:\s+(\d+)\s+\[([^\]]+)\]/ =~ l
    nid = $1.to_i
    mode = $2.to_sym
  elsif /^name:\s+(.+)$/ =~ l
    name = $1.split(/:\s+/, 2)
    case name[0]
      when 'audio output'
        name = name[0]
        wclass = Output
      when 'audio input'
        name = name[0]
        wclass = Input
      when 'audio selector'
        name = name[0]
        wclass = Selector
      when 'audio mixer'
        name = name[0]
        wclass = Mixer
      when 'beep widget'
        name = name[0]
        wclass = Beep
      when 'power widget'
        name = name[0]
        wclass = Power
      when 'pin'
        name = name[1]
        wclass = Pin
      when 'vendor widget'
        name = name[0]
        wclass = Vendor
      when 'volume widget'
        name = name[0]
        wclass = Volume
    else
      p name
      raise RuntimeError, 'Unknown widget class!'
    end
  elsif /^connections:\s+(\d+)$/ =~ l
    conns = []
  elsif /^\+\s+<\-\s+nid=(\d+)\s+/ =~ l
    conns << $1.to_i
  elsif /^(Out|In)put amp:\s+/ =~ l
    ampdir = $1
    if ampdir == "Out"
      ampdir = CTL_OUT
    elsif ampdir == "In"
      ampdir = CTL_IN
    else
      ampdir = nil
    end
  elsif /^mute=(\d+)\s+step=(\d+)\s+size=(\d+)\s+offset=(\d+)\s*$/ =~ l
    if ampdir == CTL_IN
      inamp = Amp.new(nid, 0, ampdir, $1.to_i, $2.to_i, $3.to_i, $4.to_i)
      if inamp.off > inamp.step
        p inamp
      end
    elsif ampdir == CTL_OUT
      outamp = Amp.new(nid, 0, ampdir, $1.to_i, $2.to_i, $3.to_i, $4.to_i)
      if outamp.off > outamp.step
        p outamp
      end
    end
  elsif /OUT|IN/ =~ l && /amp:|DUMP/ !~ l
    pcap = l.gsub(/[^A-Z\s]/, '').split(/\s+/)
    pindir = 0
    if pcap.include?('OUT')
      pindir |= PIN_OUT
    end
    if pcap.include?('IN')
      pindir |= PIN_IN
    end
  elsif /^$/ =~ l && nid
    w = wclass.new(name, nid, mode, conns)
    if wclass == Pin
      w.pindir = pindir
    else
      w.pindir = 0
    end
    nodes << w
    if outamp
      amps << outamp
    end
    if inamp
      amps << inamp
      if wclass == Mixer || wclass == Selector
        1.upto(conns.size - 1) do |index|
          amps << Amp.new(nid, index, inamp.dir,
            inamp.mute, inamp.step, inamp.size, inamp.off)
        end
      end
    end
    nid = mode = name = wclass = conns = inamp = outamp = ampdir = pindir = nil
  elsif /^\s*(Vendor|Device|Revision|Stepping|PCI Subvendor):\s+(0x[0-9a-fA-F]+)\s*$/ =~ l
    hdac = HDAC.new() if hdac.nil?
    tag = $1.dup()
    v = $2.hex()
    tag.downcase!
    tag.gsub!(/\s/, '_')
    hdac.instance_eval %[
      @#{tag} = #{v}
    ]
  elsif /DUMPING HDA AUDIO/ =~ l
    break
  end
end

def hdac_vendor_patch(codec, nodes, amps)
  case codec
    when 0x10ec0260# => 'Realtek ALC260',
      nodes.delete_if do |w| w.is_a?(Input) && w.nid != 5 end
    when 0x10ec0861# => 'Realtek ALC861',
    when 0x10ec0880# => 'Realtek ALC880',
      nodes.delete_if do |w| w.is_a?(Input) && w.nid != 9 && w.nid != 29 end
      nodes.collect! do |w|
        if w.nid == 29
          nw = Beep.new(w.str, w.nid, w.mode, w.conns)
          nw.pindir = w.pindir
          nw
        else
          w
        end
      end
    when 0x10ec0882# => 'Realtek ALC882',
    when 0x11d41981# => 'Analog Device AD1981HD',
    when 0x11d41983# => 'Analog Device AD1983',
    when 0x11d41986# => 'Analog Device AD1986A',
      nodes.delete_if do |w| w.is_a?(Output) && w.nid != 3 end
    when 0x434d4980# => 'CMedia CMI9880',
    when 0x83847680# => 'Sigmatel STAC9221',
      nodes.delete_if do |w| w.is_a?(Output) && w.nid != 2 end
    when 0x83847683# => 'Sigmatel STAC9221D',
      nodes.delete_if do |w|
        w.is_a?(Input) && w.nid != 6
      end
    when 0x83847690# => 'Sigmatel STAC9220'
      nodes.each do |w|
        if w.nid == 12 || w.nid == 15
          w.ctlflags |= 1 << SOUND_MIXER_LINE
        end
      end
      amps.each do |a|
        if a.nid == 20 && a.index == 0
          a.flags |= 1 << SOUND_MIXER_LINE
        end
      end
  end
end

printf("Parsing HDA dump: %s\n", dumpfile)
printf("Nodes size: %d, Amp size: %d\n", nodes.size(), amps.size())
codec = 0x00000000
unless hdac.nil?
  printf("       Vendor: 0x%08x\n", hdac.vendor)
  printf("       Device: 0x%08x\n", hdac.device)
  printf("     Revision: 0x%08x\n", hdac.revision)
  printf("     Stepping: 0x%08x\n", hdac.stepping)
  printf("PCI Subvendor: 0x%08x\n", hdac.pci_subvendor)
  codec = ((hdac.vendor & 0xffff) << 16) | (hdac.device & 0xffff)
  printf("        Codec: 0x%08x [%s]\n", codec, HDA_CODECS[codec])
end
printf("\n")

def get_recurse_output(nodes, nid, depth)
  ret = false
  if depth > MAXDEPTH
    return false
  end
  w = hdac_widget_get(nodes, nid)
  return false if w.nil?
  if w.is_a?(Output)
    w.parsed |= OUT_PATH
    ret = true
  elsif w.is_a?(Selector) || w.is_a?(Mixer)
    w.conns.each_with_index do |cnid, idx|
      if get_recurse_output(nodes, cnid, depth + 1)
        if w.selconn == -1
          w.selconn = idx
        end
        ret = true
        w.parsed |= OUT_PATH
      end
    end
  else
    ret = false
  end
  return ret
end

def get_amp_oss
  @idx ||= 0
  while @idx < SOUND_MIXER_NRDEVICES
    case @idx
      when SOUND_MIXER_VOLUME
      when SOUND_MIXER_PCM
      when SOUND_MIXER_CD
      when SOUND_MIXER_LINE
      when SOUND_MIXER_MIC
      when SOUND_MIXER_SPEAKER
      when SOUND_MIXER_BASS
      when SOUND_MIXER_TREBLE
      when SOUND_MIXER_RECLEV
    else
      @idx += 1
      return @idx - 1
    end
    @idx += 1
  end
  return -1
end

def set_outamp_flags(amps, nodes, nid, pnid, idx, depth, strategy)
  raise RuntimeError, 'Too depth' if depth > MAXDEPTH
  w = hdac_widget_get(nodes, nid)
  pw = hdac_widget_get(nodes, pnid)
  return 0 if w.nil?
  if w.is_a?(Mixer) || w.is_a?(Selector)
    fl = 0
    w.conns.each_with_index do |cnid, cidx|
      fl |= set_outamp_flags(amps, nodes, cnid, nid, cidx, depth + 1, strategy)
    end
    w.ctlflags |= fl
    return fl
  elsif w.is_a?(Output) && (w.parsed & OUT_PATH) != 0
    fl = 0
    amps.each do |a|
      if ((a.nid == pnid && a.index == idx && ((a.dir & CTL_IN) != 0)) || \
          (a.nid == pnid && pw != nil && pw.is_a?(Selector) && (pw.nconns < 2 || pw.selconn == idx || pw.selconn == -1) && ((a.dir & CTL_OUT) != 0)) || \
          (strategy == :DIRECT && a.nid == nid) || a.nid == nid) && \
          (a.flags & ~(1 << SOUND_MIXER_VOLUME)) == 0
        if pw != nil && pw.selconn == -1
          STDERR.printf("nid: %d Connection select none, set: %d for \"pcm\"\n", pw.nid, idx)
          pw.selconn = idx
        end
        # V 2
        if (w.ctlflags & (1 << SOUND_MIXER_PCM)) == 0 || \
            (pw != nil && (pw.ctlflags & (1 << SOUND_MIXER_PCM)) == 0)
          fl |= 1 << SOUND_MIXER_VOLUME
          fl |= 1 << SOUND_MIXER_PCM
          a.flags |= fl
          w.ctlflags |= fl
          pw.ctlflags |= (1 << SOUND_MIXER_VOLUME) | (1 << SOUND_MIXER_PCM) if pw != nil
        end
=begin
        # V 1
        fl |= 1 << SOUND_MIXER_VOLUME
        fl |= 1 << SOUND_MIXER_PCM
        a.flags |= fl
=end
      end
    end
    w.ctlflags |= fl
    return fl
  elsif w.is_a?(Pin) && (w.parsed & IN_PATH) != 0 && (w.pindir & PIN_IN) != 0
    rfl = 0
    amps.each do |a|
      if ((a.nid == pnid && a.index == idx && ((a.dir & CTL_IN) != 0)) || \
          (a.nid == pnid && pw != nil && pw.is_a?(Selector) && (pw.nconns < 2 || pw.selconn == idx || pw.selconn == -1) && ((a.dir & CTL_OUT) != 0)) || \
          (strategy == :DIRECT && a.nid == nid)) && \
          (a.flags & ~(1 << SOUND_MIXER_VOLUME)) == 0
        if pw != nil && pw.selconn == -1
          STDERR.printf("nid: %d Connection select none, set: %d for \"pin\"\n", pw.nid, idx)
          pw.selconn = idx
        end
        case w.str
          when /^Mic in/
            fl = 1 << SOUND_MIXER_MIC
          when /^line in/
            fl = 1 << SOUND_MIXER_LINE
          when /^CD/
            fl = 1 << SOUND_MIXER_CD
        else
          fl = get_amp_oss()
          if fl < 0
            fl = 0
          else
            fl = 1 << fl
          end
        end
        if strategy == :MIXER
          fl |= 1 << SOUND_MIXER_VOLUME
        end
        a.flags |= fl
        rfl |= fl
      end
    end
    w.ctlflags |= rfl
    return rfl
  elsif w.is_a?(Beep)
    fl = 0
    amps.each do |a|
      if ((a.nid == pnid && a.index == idx && ((a.dir & CTL_IN) != 0)) || \
          (a.nid == pnid && pw != nil && pw.is_a?(Selector) && (pw.nconns < 2 || pw.selconn == idx || pw.selconn == -1) && ((a.dir & CTL_OUT) != 0)) || \
          (strategy == :DIRECT && a.nid == nid)) && \
          (a.flags & ~(1 << SOUND_MIXER_VOLUME)) == 0
        if pw != nil && pw.selconn == -1
          STDERR.printf("nid: %d Connection select none, set: %d for \"beep\"\n", pw.nid, idx)
          pw.selconn = idx
        end
        fl = 0
        fl |= 1 << SOUND_MIXER_VOLUME
        fl |= 1 << SOUND_MIXER_SPEAKER
        a.flags |= fl
      end
    end
    w.ctlflags |= fl
    return fl
  end
  return 0
end

def dump_ctl(amps, nodes, ctl)
    w = hdac_widget_get(nodes, ctl.nid)
    printf("    |\n")
    printf("    +- nid: %2d index: %2d ", ctl.nid, ctl.index)
    if (w.is_a?(Mixer) || w.is_a?(Selector)) && w.conns.size() > 0
      printf("(nid: %2d) ", w.conns[ctl.index])
    end
    printf("mute: %d step: %3d size: %3d off: %3d dir=%s ossmask=0x%08x\n",
      ctl.mute, ctl.step, ctl.size, ctl.off, ctl.dir.to_s, ctl.flags)
end

hdac_vendor_patch(codec, nodes, amps)

nodes.delete_if do |w| w.mode == :DIGITAL end

nodes.delete_if do |w|
  w.is_a?(Pin) && /none/ =~ w.str
end

amps.delete_if do |a| hdac_widget_get(nodes, a.nid).nil? end

assoc = []
strategy = :MIXER
found = false
nodes.each do |w|
  if w.is_a?(Pin) && (w.pindir & PIN_OUT) != 0 && /headphone|line out|speaker/ =~ w.str
    w.conns.each_with_index do |cnid, idx|
      cw = hdac_widget_get(nodes, cnid)
      next if cw.nil?
      if cw.is_a?(Mixer) || cw.is_a?(Selector)
        if get_recurse_output(nodes, cnid, 1)
          if w.selconn == -1
            w.selconn = idx
          end
          w.parsed |= OUT_PATH
          found = true
        end
      end
    end
  end
end

unless found
  strategy = :DIRECT
  nodes.each do |w|
    if w.is_a?(Pin) && (w.pindir & PIN_OUT) != 0 && /headphone|line out|speaker/ =~ w.str
      w.conns.each_with_index do |cnid, idx|
        cw = hdac_widget_get(nodes, cnid)
        next if cw.nil?
        if get_recurse_output(nodes, cnid, 1)
          if w.selconn == -1
            w.selconn = idx
          end
          cw.parsed |= OUT_PATH
          w.parsed |= OUT_PATH
          found = true
        end
      end
    end
  end
end

raise RuntimeError, 'No output found!' unless found
def get_recurse_input(nodes, nid, depth)
  if depth > MAXDEPTH
    return false
  end
  w = hdac_widget_get(nodes, nid)
  return false if w.nil?
  if w.is_a?(Input) || w.is_a?(Selector) || w.is_a?(Mixer)
    ret = false
    w.conns.each_with_index do |cnid, cidx|
      if get_recurse_input(nodes, cnid, depth + 1)
        if w.selconn == -1
          w.selconn = cidx
        end
        w.parsed |= IN_PATH
        ret = true
      end
    end
    return ret
  elsif w.is_a?(Pin) && (w.pindir & PIN_IN) != 0
    if /CD|line in|Mic in/ =~ w.str
      w.parsed |= IN_PATH
      return true
    end
  #elsif w.is_a?(Mixer)
  #  if (w.parsed & OUT_PATH) != 0
  #    w.parsed |= IN_PATH
  #    return true
  #  end
  end
  return false
end

def set_recselector(nodes, nid)
  w = hdac_widget_get(nodes, nid)
  chcnt = 0
  return false if w.nil?
  #return false unless (w.parsed & IN_PATH) != 0
  #return false unless w.is_a?(Input) || w.is_a?(Selector)
  w.conns.each do |cnid|
    cw = hdac_widget_get(nodes, cnid)
    next if cw.nil?
    chcnt += 1
  end
  if chcnt > 1
    w.parsed |= REC_SELECTOR
    return true
  else
    w.conns.each do |cnid|
      if set_recselector(nodes, cnid)
        return true
      end
    end
  end
  return false
end

nodes.each do |w|
  if w.is_a?(Input)
    get_recurse_input(nodes, w.nid, 0)
  end
end

nodes.each do |w|
  if (strategy == :MIXER && ((w.is_a?(Mixer) || w.is_a?(Selector)) && (w.parsed & OUT_PATH) != 0)) || \
      (strategy == :DIRECT && (w.parsed & (OUT_PATH|IN_PATH)) != 0)
    w.ctlflags |= set_outamp_flags(amps, nodes, w.nid, -1, 0, 0, strategy)
  elsif w.is_a?(Beep)
    amps.each do |a|
      if a.nid == w.nid
        a.flags |= 1 << SOUND_MIXER_VOLUME
        a.flags |= 1 << SOUND_MIXER_SPEAKER
        w.ctlflags |= 1 << SOUND_MIXER_VOLUME
        w.ctlflags |= 1 << SOUND_MIXER_SPEAKER
      end
    end
  end
end

def set_inamp_flags(amps, nodes, nid)
  w = hdac_widget_get(nodes, nid)
  return 0 if w.nil? #|| (w.parsed & IN_PATH) == 0
  if w.is_a?(Input) || w.is_a?(Selector)
    amps.each do |a|
      if a.nid == nid
        a.flags |= 1 << SOUND_MIXER_RECLEV
        w.ctlflags |= 1 << SOUND_MIXER_RECLEV
        return (1 << SOUND_MIXER_RECLEV)
      end
    end
    w.conns.each do |cnid|
      cw = hdac_widget_get(nodes, cnid)
      next if cw.nil?
      if cw.is_a?(Selector)
        fl = set_inamp_flags(amps, nodes, cnid)
        if fl != 0
          cw.ctlflags |= fl
          w.ctlflags |= fl
          return fl
        end
      end
    end
  end
  return 0
end

nodes.each do |w|
  if w.is_a?(Input) && (w.parsed & IN_PATH) != 0
    set_inamp_flags(amps, nodes, w.nid)
    set_recselector(nodes, w.nid)
  end
end

#nodes.delete_if do |w| w.parsed == 0 && w.ctlflags == 0 end
#amps.delete_if do |a| a.flags == 0 || hdac_widget_get(nodes, a.nid).nil? end

[
  [1 << SOUND_MIXER_VOLUME, 'Master Volume (OSS: %s)', 'vol'],
  [1 << SOUND_MIXER_PCM, 'PCM Volume (OSS: %s)', 'pcm'],
  [1 << SOUND_MIXER_SPEAKER, 'Speaker/Beep (OSS: %s)', 'speaker'],
  [1 << SOUND_MIXER_CD, 'CD Volume (OSS: %s)', 'cd'],
  [1 << SOUND_MIXER_MIC, 'Microphone Volume (OSS: %s)', 'mic'],
  [1 << SOUND_MIXER_LINE, 'Line-in Volume (OSS: %s)', 'line'],
  [1 << SOUND_MIXER_RECLEV, 'Recording Level (OSS: %s)', 'rec']
].each do |fl, str, dev|
  printed = false
  amps.each do |a|
    if (a.flags & fl) != 0 && !printed
      printed = true
      pstr = sprintf(str, dev)
      printf("  %s\n", pstr);
      cnt = 0
      amps.each do |aa|
        if (aa.flags & fl) != 0
          dump_ctl(amps, nodes, aa)
          cnt += 1
        end
      end
      printf("\n")
      break
    end
  end
end

idx = 0
while idx < SOUND_MIXER_NRDEVICES
  case idx
    when SOUND_MIXER_VOLUME
    when SOUND_MIXER_PCM
    when SOUND_MIXER_CD
    when SOUND_MIXER_LINE
    when SOUND_MIXER_MIC
    when SOUND_MIXER_SPEAKER
    when SOUND_MIXER_BASS
    when SOUND_MIXER_TREBLE
    when SOUND_MIXER_RECLEV
  else
    printed = false
    amps.each do |a|
      if (a.flags & (1 << idx)) != 0
        printf("  Unknown Ctl (OSS: %s)\n", OSS_LABELS[idx]) unless printed
        dump_ctl(amps, nodes, a)
      end
    end
  end
  idx += 1
end

def flagstr(fl)
  ret = ''
  0.upto(SOUND_MIXER_NRDEVICES - 1) do |i|
    if (fl & (1 << i)) != 0
      ret << ' '
      ret << OSS_LABELS[i]
    end
  end
  ret.sub(/^\s+/, '')
  ret
end

def dump_dac(nodes, nid, depth)
  return false if depth > MAXDEPTH
  w = hdac_widget_get(nodes, nid)
  return false if w.nil? || (w.parsed & OUT_PATH) == 0
  unless w.is_a?(Pin)
    printf("                 ^\n");
    printf("                 |\n");
    printf("             ")
  end
  printf("  nid=%s [%s]\n", w.nid, w.str)
  if w.is_a?(Pin)
    printf("    ^\n")
    printf("    |\n")
    printf("    +-----<------+\n")
  end
  if w.is_a?(Output)
    printf("\n")
    return true
  elsif w.is_a?(Mixer)
    w.conns.each do |cnid|
      next if hdac_widget_get(nodes, cnid).is_a?(Pin)
      if dump_dac(nodes, cnid, depth + 1)
        return true
      end
    end
  elsif w.is_a?(Selector) || w.is_a?(Pin)
    if dump_dac(nodes, w.conns[w.selconn], depth + 1)
      return true
    end
  end
  return false
end

printed = false
nodes.each do |w|
  if w.is_a?(Pin) && (w.parsed & OUT_PATH) != 0
    printf("\n\nPlayback path:\n\n") unless printed
    printed = true
    dump_dac(nodes, w.nid, 0)
  end
end

printed  = false
nodes.each do |w|
  if (w.parsed & REC_SELECTOR) != 0
    printf("\nRecording path:\n") unless printed
    printed = true
    printf("\n    nid=%d (%s) %s\n", w.nid, w.str, flagstr(w.ctlflags))
    w.conns.each do |cnid|
      cw = hdac_widget_get(nodes, cnid)
      next if cw.nil?
      printf("     |\n")
      printf("     +<- nid=%d (%s) %s\n", cw.nid, cw.str, flagstr(cw.ctlflags))
    end
  end
  #if w.selconn == -1
  #  p [w.nid, w.selconn, w.conns, w.class]
  #end
end

#nodes.each do |w|
#  printf("pcm0:             nid: %d [ANALOG]\n", w.nid)
#  printf("pcm0:     Parse flags: 0x%08x\n", w.parsed)
#end
